# -* perl *-

## progressive multiple alignment

use MLocarna;
use strict;

##------------------------------------------------------------
## options
use Getopt::Long;
use Pod::Usage;

my $help;
my $man;

my $quiet;
my $verbose;

my $match;
my $mismatch;
my $indel;
my $struct_weight;

my $struct_local;
my $sequ_local;

my $width;
my $tau;

my $min_prob=0.01;
my $tree_min_prob=0.05;
my $exp_prob;

my $max_diff_am;

my $no_lp;

my $tree_file;
my $tgtdir;

my $mfasta_file;

my $skip_pp=0; ## skip the computation of pair probabilities for files that exist already 

my $ext_library;

Getopt::Long::Configure("no_ignore_case");

GetOptions(
	   "match|m=i" => \$match,
	   "mismatch|M=i"=> \$mismatch,
	   "indel" => \$indel,
	   "struct-weight=i" => \$struct_weight,
	   "tau=f" => \$tau,
	   "min-prob|p=f" => \$min_prob,
	   "tree-min-prob|P=f" => \$tree_min_prob,
	   "exp-prob=f" => \$exp_prob,
	   "noLP|l" => \$no_lp,
	   "struct-local=s" => \$struct_local,
	   "sequ-local=s" => \$sequ_local,
	   "width=i" => \$width,
	   "treefile=s" => \$tree_file,
	   "outdir|tgtdir=s" => \$tgtdir,
	   "max-diff-am=i" => \$max_diff_am,

	   "extlib" => \$ext_library,

	   "skip-pp" => \$skip_pp,
	   
	   "verbose" => \$verbose,
	   "quiet" => \$quiet,
	   
	   "help"=> \$help,
	   "man" => \$man
	   ) || pod2usage(2);

pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

## construct parameter string for locarna

my $locarna_params="";
my $locarna_params_tree="";

$locarna_params.=" -v" if ($verbose);
$locarna_params.=" --match $match" if (defined($match));
$locarna_params.=" --mismatch $mismatch" if (defined($mismatch));
$locarna_params.=" --indel $indel" if (defined($indel));
$locarna_params.=" --struct-weight $struct_weight" if (defined($struct_weight));
$locarna_params.=" --width $width" if (defined($width));
$locarna_params.=" --sequ-local $sequ_local" if (defined($sequ_local));
$locarna_params.=" --struct-local $struct_local" if (defined($struct_local));

$locarna_params.=" --max-diff-am $max_diff_am" if (defined($max_diff_am));

$locarna_params_tree=$locarna_params;

## special treatment of 

$locarna_params.=" --min-prob $min_prob" if (defined($min_prob));
if (defined($tree_min_prob)) {
    $locarna_params.=" --min-prob $tree_min_prob";
} else {
    $locarna_params_tree.=" --min-prob $min_prob" if (defined($min_prob));
}



## ----------------------------------------
# mfasta file

if ($#ARGV == 0) {
    $mfasta_file = $ARGV[0];
}else {
    pod2usage(1);
}



## target/output directory
if (!defined($tgtdir)) {
    $tgtdir = $mfasta_file;
    $tgtdir =~ s/\.mfasta|\.fasta//;
    $tgtdir .= ".out";
}


if ($verbose) {
    print "Locarna Parameter for progressive alignment: $locarna_params\n";
    print "Locarna Parameter for guide tree construction: $locarna_params_tree\n";
}



##------------------------------------------------------------



my $intermediate_name_base="intermediate";
my $intermediate_name_counter=0;

sub new_intermediate_name {
    $intermediate_name_counter++;
    return "$intermediate_name_base$intermediate_name_counter.pp";
}

my $node_sym="\$\$nodesym";

sub parse_tree;
sub parse_dp_ps;

## ---------------------------------------------------------------------------
# main program

if (! -e $tgtdir) {
    mkdir $tgtdir || die "Cannot create $tgtdir";
} else {
    print STDERR "Warning: $tgtdir exists already.\n" if ($verbose);
}

my %mfasta = parse_mfasta($mfasta_file);

my %filenames;

my $tree;

if (defined($tree_file)) {
    open(TREE,"$tree_file") || die "Cannot read $tree_file\n";
}

print "Compute pair probs ...\n";

chdir $tgtdir;

for my $name (keys %mfasta) {
    my $RNAfold_args="";
    $RNAfold_args=">/dev/null" unless $verbose;
    
    if ( $skip_pp && -e $name ) {
	# skip
    } else {
	system "printf '>$name\n$mfasta{$name}' | RNAfold -p $RNAfold_args";
    
	my $shortname=substr $name,0,12;
	unlink "$shortname\_ss.ps";
	
	rename "$shortname\_dp.ps", "$name"
	    || die "Cannot rename $shortname\_dp.ps to $name\n";
    }
}

if (defined($tree_file)) {
    $tree=<TREE>;
    chomp($tree);
    close TREE;
} else {    
    # --------------------------------------------------
    # compute score matrix
    #

    my @score_matrix;
    my @pairwise_alns;
    
    my @names = keys %mfasta;

    ## compute pairwise scores in score_matrix
    
    print "Compute pairwise alignments ... \n" unless $quiet;

    my $num=0;
    for (my $a=0; $a<=$#names; $a++) {
	for (my $b=0; $b<$a; $b++) {
	    $num++;
	    printf "Align %d/%d : $names[$a] and %d/%d: $names[$b] (%d/%d)\n",$a+1,$#names+1,$b+1,$a,$num,($#names*($#names+1))/2 if $verbose;
	    my @aln = compute_alignment_from_dps($names[$a],
						 $names[$b],
						 $locarna_params_tree); 
	    
	    
	    $pairwise_alns[$a][$b] = [ @aln ];
	    }
    }
    
    for (my $a=0; $a<=$#names; $a++) {
	for (my $b=0; $b<$a; $b++) {
	    
	    my @aln = @{ $pairwise_alns[$a][$b] };
	    
	    # print "Alig. of $names[$a] --- $names[$b]:\n@aln\n";
	    
	    $aln[0] =~ /Score: ([\d\.\-]+)/ || die "Cannot extract score from $aln[0].\n";
	    my $score=$1;
	    
	    $score_matrix[$a][$b] = $score;
	    $score_matrix[$b][$a] = $score;
	    
	}
    }

    open(MAT,">result.matrix") || die "Cannot write score matrix\n";
    for (my $a=0; $a<=$#names; $a++) {
	for (my $b=0; $b<=$#names; $b++) {
	    if ($a==$b) {print MAT "     0 ";} else { 
		printf MAT "%6d ",$score_matrix[$a][$b];
	    }
	}
	print MAT "\n";
    }
    close MAT;
    # system("cat $tgtdir/matrix");
    

    if ($ext_library) {
	## ------------------------------------------------------------
	## compute "extended pairprobability library"
	#

	my @pairprobs;
	my @sequences;

	for (my $a=0; $a<=$#names; $a++) {
	    my $seqA;
	    my $pairprobsA_ref;
	    ($seqA,$pairprobsA_ref) = parse_dp_ps("$names[$a]");
	    
	    $sequences[$a] = $seqA;
	    $pairprobs[$a] = [ @{ $pairprobsA_ref } ];
	}
	
	print "Compute extended pairprobability library ...\n";
	
	## make a deep copy of pair probabilities
	#
	my @ext_pairprobs;
	for (my $a=0; $a<=$#names; $a++) {
	    my @eppA;
	    my @ppA = @{ $pairprobs[$a] };
	    for my $i ( 0..$#ppA ) {
		if (defined $ppA[$i]) {
		    my @ppAa = @{ $ppA[$i] };
		    for my $j ( 0..$#ppAa ) {
			if (defined $ppAa[$j]) {
			    $eppA[$i][$j] = $ppA[$i][$j];
			}
		    }
		}
	    }
	    $ext_pairprobs[$a] = [ @eppA ];
	}
	
	for (my $a=0; $a<=$#names; $a++) {
	    print "Compute ext lib for $names[$a]\n";
	    
	    my @ext_pairprobsA = @{ $ext_pairprobs[$a] }; # we accummulate the extension!
	    
	    for (my $b=0; $b<=$#names; $b++) {
		if ($a != $b) {
		    
		    my $aliA; # alignment-string for sequence a
		    my $aliB; # alignment-string for sequence b
		    
		    if ($a>$b) {
			my @alnAB = @{ $pairwise_alns[$a][$b] };
			($aliA,$aliB) = parse_alignment($names[$a],$names[$b],\@alnAB);
		    } else {
			my @alnBA = @{ $pairwise_alns[$b][$a] };
			($aliB,$aliA) = parse_alignment($names[$b],$names[$a],\@alnBA);
		    }
		    
		    my @sp2apA = seqpos_to_alipos($aliA);
		    my @sp2apB = seqpos_to_alipos($aliB);
		    
		    my @ap2spA = alipos_to_seqpos($aliA);
		    my @ap2spB = alipos_to_seqpos($aliB);
		    
		    my $lengthA=$#sp2apA+1;
		    my $lengthB=$#sp2apB+1;
		    
		    print "  $a:$names[$a] ($lengthA) --- $b:$names[$b]\n";
		    
		    my @pairprobsB = @{ $pairprobs[$b] };
		    
		    for (my $iA=0; $iA<$lengthA; $iA++) {
			for (my $jA=$iA+1; $jA<$lengthA; $jA++) {
			    my $pA = (defined $ext_pairprobsA[$iA][$jA])?$ext_pairprobsA[$iA][$jA]:$min_prob;
			    if ($pA > $min_prob) {
				my $iB = $ap2spB[$sp2apA[$iA]];
				my $jB = $ap2spB[$sp2apA[$jA]];
				
				if ($iB != -1 && $jB != -1) {
				    # print "$iA,$jA <--> $sp2apA[$iA],$sp2apA[$jA] <--> $iB,$jB\n";
				    my $pB = (defined $pairprobsB[$iB][$jB])?$pairprobsB[$iB][$jB]:$min_prob;
				    if ($pB>$min_prob) {
					my $new_pA = exp( log($pA) * (1/(1+$pB)));
					$ext_pairprobsA[$iA][$jA] = $new_pA;
					my @orig_pairprobsA=@{ $pairprobs[$a] };
					my $orig_pA  = $orig_pairprobsA[$iA][$jA];
					print "Enhance ($iA,$jA):$pA by ($iA,$jB):$pB --> $new_pA/$orig_pA\n";
				    }
				}
			    }
			}
		    }
		}
	    }
	}
	
	## --------------------------------------------------
	## write pp files with extended probabilities
	#
	
	for (my $a=0; $a<=$#names; $a++) {
	    # print "write extended probs for $names[$a]\n";

	    my $seq=$sequences[$a];
	    my $len=length($seq);
	    my @ext_pairprobs = @{ $ext_pairprobs[$a] }; # the extended probs
	    

	    open(PP_OUT, ">$names[$a].pp") || die "Cannot write to $names[$a].pp\n"; 
	    
	    print PP_OUT "$names[$a] $seq\n\n#\n";
	    
	    for (my $i=0; $i<$len; $i++) {
		for (my $j=$i+1; $j<$len; $j++) {
		    if (defined $ext_pairprobs[$i][$j]) {
			my $p = $ext_pairprobs[$i][$j];
			if ($p > $min_prob) {
			    print PP_OUT "$i $j $p\n";
			}
		    }
		}
	    }
	}
	
	#print "\n\n";
	

	for (my $a=0; $a<=$#names; $a++) {
	    rename "$names[$a].pp", "$names[$a]";
	}
	
	#
	##
	## --------------------------------------------------
    }


    $tree = compute_tree(\%mfasta,\@score_matrix);    

    open(TREE,">result.tree")|| die "Cannot write to $tgtdir/result.tree";
    print TREE "$tree\n";
    close TREE;
}

my @tree_postorder = parse_tree($tree);

# print "TREE: <@tree_postorder>\n";
    

# print "@tree_postorder\n";

## if we read the tree from a file, then we have to shorten the names,
## and substitute special characters
if (defined($tree_file)) {
    for (my $i=0; $i<=$#tree_postorder; $i++) {
	if ( $tree_postorder[$i] ne $node_sym ) {
	    $tree_postorder[$i] =  substr $tree_postorder[$i],0,16;
	    $tree_postorder[$i] =~ s/[^a-zA-Z\d]/_/g;
	}
    }
}


my @stack=();

for my $item (@tree_postorder) {   
    if ($item eq $node_sym) {
	# align two topmost stack entrys and push result
	my $op1 = $stack[$#stack-1];
	my $op2 = $stack[$#stack];
	
	$#stack -= 2;
	
	my $intermediate_name = new_intermediate_name();
	(my $intermediate_name_aln=$intermediate_name) =~ s/.pp/.aln/;

	print "--------------------\nAlign $op1 with $op2 --> $intermediate_name\n\n" unless $quiet;
	system "locarna $op1 $op2 $locarna_params --pp=$intermediate_name --clustal=$intermediate_name_aln >/dev/null"
	    || die "Cannot execute locarna\n";
	
	system "cat $intermediate_name_aln" unless ($quiet);
	
	push @stack, $intermediate_name;
	
    } else {
	push @stack, $item;
    }
}

if ($#stack!=0) {
    print STDERR "Strange error occured! Exit.\n";
    print "STACK: @stack\n";
    exit -1;
}


(my $intermediate_name_aln=$stack[$#stack]) =~ s/.pp/.aln/;
system "cp", "$intermediate_name_aln", "result.aln";

#open(IN, "$stack[$#stack]") || die "Cannot read $tgtdir/$stack[$#stack]\n";
#open(OUT, ">result.aln") || die "Cannot write to $tgtdir/result.aln\n";
#print OUT "CLUSTAL --- created by mlocarna\n\n";
#while ((my $line=<IN>) !~ "^#") {print OUT $line;}
#close OUT;
#close IN;

system "cp", "$stack[$#stack]", "result.pp";

sub parse_tree {
    my ($tree) = @_;

    my @list;
    
    $tree =~ s/:[\d.e-]+//g;
    
    print "Tree: $tree\n" unless $quiet;
    
    my $brcount=0;
    for (my $i=0; $i< length $tree; $i++) {
	my $c=substr $tree,$i,1;
	
	if ($c eq "(") {
	    $brcount++;
	} elsif ($c eq ")") {
	    $brcount--;
	    if ($brcount<0) {
		print STDERR "Parse error in tree.";
		exit -1;
	    }
	    
	    push @list, $node_sym;
	} elsif ($c eq ",") {
	    ## ignore, although we could do syntax checking
	} else {
	    my $rest=substr $tree,$i;
	    $rest =~ /^([^(),]+)/;
	    my $token = $1;
	    $i+=(length $token)-1;
	    push @list, $token;
	}
    }

    return @list;
}

## compute a WPGMA-tree from the given mfasta-hash
sub compute_tree {
    # print "Compute WPGMA-Tree ...\n";

    my ($mfasta_ref, $score_matrix_ref) = @_;
    
    my %mfasta = %$mfasta_ref;
    my @score_matrix = @$score_matrix_ref;
    
    my @names = keys %mfasta;

    
    ## compute tree by wpgma applied to score matrix
    
    my @clusters; # a list of the clusters
    my @trees; # a list of sub-trees
    my @cluster_sizes;

    for (my $i=0; $i<=$#names; $i++) {
	$clusters[$i]=$i;
	$trees[$i]=$names[$i];
	$cluster_sizes[$i]=1;
    }
    
    
    my $NEG_INFINITY=-1e10;
    
    while ($#clusters>0) {
	
	#print "@clusters\n";
	#print "@trees\n";
	#print "@cluster_sizes\n";
	
        ## find the nearest two clusters
	
	my $max_i;
	my $max_j;
	my $max_score=$NEG_INFINITY;
	for (my $i=0; $i<=$#clusters; $i++) {
	    for (my $j=$i+1; $j<=$#clusters; $j++) {
		my $score=$score_matrix[$clusters[$i]][$clusters[$j]];
		if ($score > $max_score) {
		    $max_i=$i; 
		    $max_j=$j;
		    $max_score=$score;
		}
	    }
	}
	
	print "Merge $clusters[$max_i] $clusters[$max_j] --> $clusters[$max_i]\n" if $verbose;
	
	## recompute distances
	my $cluster_i = $clusters[$max_i];
	my $cluster_j = $clusters[$max_j];

	$clusters[$max_j] = $clusters[$#clusters];
	$clusters[$max_i] = $clusters[0];
	$clusters[0] = $cluster_i;
	
	for (my $i=1; $i<$#clusters; $i++) {
	    $score_matrix[$clusters[0]][$clusters[$i]]
		= ($cluster_sizes[$cluster_i] * $score_matrix[$cluster_i][$clusters[$i]] + $cluster_sizes[$cluster_j] * $score_matrix[$cluster_j][$clusters[$i]]) / ($cluster_sizes[$cluster_i]+$cluster_sizes[$cluster_j]);
	    $score_matrix[$clusters[$i]][$clusters[0]] = 
		$score_matrix[$clusters[0]][$clusters[$i]];
	}
	
	my $new_tree="($trees[$cluster_i],$trees[$cluster_j])";
	print "Merge $trees[$cluster_i] $trees[$cluster_j] --> $new_tree\n" if $verbose;
	
	$trees[$clusters[0]]=$new_tree;
	
	#$trees[$clusters[$#clusters]]="undef";
	#$cluster_sizes[$clusters[$#clusters]]=0;
    
	$cluster_sizes[$clusters[0]]=$cluster_sizes[$cluster_i]+$cluster_sizes[$cluster_j]; 
	
	$#clusters--;
    }


    print "Computed-tree: $trees[$clusters[0]]\n" if $verbose;

    return $trees[$clusters[0]];
}



sub seqpos_to_alipos {
    my ($ali) = @_;
    
    my @res;

    my $seqpos=0;
    my $alipos=0;
    
    for my $c (split //, $ali) {
	if ($c =~ /[ACGUT]/) {
	    push @res,$alipos;
	    $seqpos++;
	}
	$alipos++;
    }

    return @res;
}

sub alipos_to_seqpos {
    my ($ali) = @_;
    
    my @res;

    my $seqpos=0;
    my $alipos=0;
    
    for my $c (split //, $ali) {
	if ($c =~ /[ACGUT]/) {
	    push @res,$seqpos;
	    $seqpos++;
	} else {
	    push @res,-1;
	}
	$alipos++;
    }

    return @res;
}


sub parse_dp_ps {
    my ($filename) = @_;
    local *IN;
    
    open(IN,$filename) || die "Cannot read $filename for parsing as dp-ps file.\n";
    
    my $seq="";
    my @pairprobs;
    
    while (my $line=<IN>) {
	if ($line =~ /^\/sequence \{ \(/) {
	    while (defined($line = <IN>) && ($line !~  /\} def/  ))  {
		chomp $line;
		$line =~ s/\\$//;
		$seq .= $line;
	    }
	    print "parse_dp_ps $filename: $seq\n";
	}
	
	if ($line =~ /(\d+)\s+(\d+)\s+(\S+)\s+ubox/) {
	    $pairprobs[$1-1][$2-1]=$3*$3;
	}
    }
    
    close IN;
    
    $seq ne "" || die "Empty sequence in dp.ps file $filename\n";
    
    return ($seq,\@pairprobs);
}


__END__

=head1 NAME

MLocarna - local multiple alignment of RNA

=head1 SYNOPSIS

mlocarna [options] <file>

 Options:
   --help            brief help message
   --man             full documentation

    -v, --verbose                 Verbose
    -q, --quiet                   Quiet
    -m, --match=<s>               Match score
    -M, --mismatch=<s>            Mismatch score
    -i, --indel=<s>               Indel score
    -s, --struct-weight=<s>       Maximal weight of 1/2 arc-match
    -t, --tau=<f>                 Tau factor in percent
    -p, --min-prob=<f>            Minimal probability
    -P, --tree-min-prob=<f>       Minimal probability for constructing the guide tree
    -e, --exp-prob=<f>            Expected probability
    -l, --noLP                    No lonely pair
    --treefile                    guide tree file
    --tgtdir                      target directory
    --max-diff-am=<int>           Maximal length difference of arc-matchs
    --struct-local=<bool>         Structure local
    --sequ-local=<bool>           Sequence local
    -w, --width=<columns>         Output width
    --skip-pp                     Skip computation of pair probs for already existing files
    --extlib                      Use library extension -- experimental feature

=head1 DESCRIPTION

B<MLocarna> computes a multiple global or local sequence-structure alignment
of the RNA molecules that are given in input file <file> in mfasta format.
Results are written to directory <dir>. If the file tree is given, contained
tree (in NEWICK-tree format) is used as guide tree for the progressive
alignment. 

Whenever parameters are not specified explicitly, we use the locarna defaults,
except for min-prob (0.01) and tree-min-prob (0.1).

=cut
