#include "aligner.hh"
#include "math.h"

using namespace std;

/*
  Note on max-diff-am heuristic:
  this heuristic avoids additional effort for the
  computation of arcs with length difference > d
  If the score is computed anyway the arc match
  is still scored
*/

// ------------------------------------------------------------
// Aligner: align / compute similarity
//

Aligner::Aligner(const Aligner &a)
    :params(a.params),
     r(a.r),
     D(a.D),
     Ms(a.Ms),
     seqA(a.seqA),
     bpsA(a.bpsA),
     seqB(a.seqB),
     bpsB(a.bpsB),
     alignment(a.alignment),
     max_i(a.max_i),
     max_j(a.max_j),
     sigma_tab(a.sigma_tab),
     D_created(a.D_created)
{}

const Aligner & Aligner::operator = (const Aligner &a) {
    params=a.params;
    r=a.r;
    D=a.D;
    Ms=a.Ms;
    seqA=a.seqA;
    bpsA=a.bpsA;
    seqB=a.seqB;
    bpsB=a.bpsB;
    alignment=a.alignment;
    max_i=a.max_i;
    max_j=a.max_j;
    sigma_tab=a.sigma_tab;
    D_created=a.D_created;
    
    return a;
}

Aligner::Aligner(Sequence &_seqA, BasePairs &_bpsA,
		 Sequence &_seqB, BasePairs &_bpsB,
		 AlignerParams *_ap    
		 )
    : params(_ap),
      seqA(_seqA), bpsA(_bpsA), seqB(_seqB), bpsB(_bpsB),
      r(1,1,_seqA.length(),_seqB.length()),
      alignment(_seqA,_seqB),D_created(false) {

    Ms.resize(params->STRUCT_LOCAL?8:1);
  
    D.resize(bpsA.num_bps(),bpsB.num_bps());
	
    for (uint k=0; k<(params->STRUCT_LOCAL?8:1); k++) {
	Ms[k].resize(seqA.length()+1,seqB.length()+1);
    }
  
    precompute_sigma();
}

/**
   returns similarity of two alignment columns
 */
score_t Aligner::sigma_(int ia, int ib) {
    AliColumn &colA=seqA[ia];
    AliColumn &colB=seqB[ib];
    
    score_t score=0;
    
    for (size_t i=0; i<colA.size(); i++) {
	for (size_t j=0; j<colB.size(); j++) {
	    if (colA[i]!='N' && colB[j]!='N') {
		score +=
		    (colA[i]==colB[j])
		    ?params->match_score
		    :params->mismatch_score;
	    }
	}
    }

    return  score / (score_t)(colA.size()*colB.size()) ;
}

void Aligner::precompute_sigma() {
    int lenA=seqA.length();
    int lenB=seqB.length();
        
    sigma_tab.resize(lenA+1,lenB+1);
    for (int i=1; i<=lenA; ++i) {
	for (int j=1; j<=lenB; ++j) {
	    sigma_tab(i,j) = sigma_(i,j);
	}
    }
}



// standard cases in alignment: base match, base in/del, arc match
// (for structure local alignment this is extended by exclusion handling)
// 
// if lonely basepairs are disallowed, there is special treatment
score_t Aligner::align_noex(ScoreMatrix &M,int al,int bl,int i, int j) {
    
    score_t max_score;
    
    // base match
    max_score = M(i-1,j-1) + sigma(i,j);
    
    if (params->ungappedA[i])
      return max_score; // allow only base-match if ungappedA[i]
 
    // base del
    max_score=max(max_score,
		  M(i-1,j)+params->indel_score);
    // base ins
    max_score=max(max_score,
		  M(i,j-1)+params->indel_score);
    
    // arc match
    
    if (params->no_lonely_pair) {
	max_score =
	    align_noex_noLP(max_score,M,al,bl,i,j);
    } else {
	// standard case for arc match (without restriction to lonely pairs)
	//
	
	const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i);
	const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j);
	
	// for all pairs of arcs in A and B that have right ends i and j, respectively
	//
	for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin();
	     arcA!=adjlA.end() && arcA->left() > al; ++arcA) {
	    for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin();
		 arcB!=adjlB.end() && arcB->left() > bl ; ++arcB) {
		//
		// consider score for match of basepairs
		max_score =
		    max(max_score,
			M(arcA->left()-1,arcB->left()-1)
			+ D(arcA->idx(),arcB->idx())
			+ arcA->weight() + arcB->weight()
			+ tau(arcA->left(),arcB->left(),i,j)
			);
	    }
	}
    }
    return max_score;
}

// special arc match case in alignment with no-lonely-pairs
score_t Aligner::align_noex_noLP(score_t max_score,ScoreMatrix &M,int al,int bl,int i, int j) {

    const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i);
    const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j);

    // for all pairs of arcs in A and B that have right ends i and j, respectively
    // and that are stacked with an enclosed arc
    //

    for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin();
	 arcA!=adjlA.end() && arcA->left() > al; ++arcA) {
	if (bpsA.exists_arc(arcA->left()+1,i-1)) {
	    const Arc &arcA2 = bpsA.arc(arcA->left()+1,i-1);
	    for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin();
		 arcB!=adjlB.end() && arcB->left() > bl ; ++arcB) {
		if (bpsB.exists_arc(arcB->left()+1,j-1)) {
		    //
		    // consider score for match of basepairs
		    // 
		    const Arc &arcB2 = bpsB.arc(arcB->left()+1,j-1);
		    max_score=
			max(max_score,
			    M(arcA->left()-1,arcB->left()-1) 
			    + D(arcA2.idx(),arcB2.idx())
			    + arcA->weight() + arcB->weight()
			    + tau(arcA->left(),arcB->left(),i,j)
			    + arcA2.weight() + arcB2.weight()
			    + tau(arcA->left()+1,arcB->left()+1,i-1,j-1));
		}
	    }
	}
    }
    return max_score;
}


void Aligner::init_state(ScoreMatrix &M, int al, int ar, int bl, int br, 
			 score_t indelA, score_t exclusion_scoreA,
			 score_t indelB, score_t exclusion_scoreB) {
    M(al,bl)=0;
    
    for (int i=al+1; i<ar; i++)
	M(i,bl) = (i-al)*indelA + exclusion_scoreA;
    
    for (int j=bl+1; j<br; j++)
	M(al,j) = (j-bl)*indelB  + exclusion_scoreB;
}


void Aligner::align_in_arcmatch(int al,int ar,int bl,int br,
				bool allow_exclusion) {

    // cout << al << " " << ar <<" " << bl << " " << br <<endl;

    // if in a sequence the state is not open than gaps with cost indel_score have to be introduced.
    // In open cases the maximal similarity is 0 in the row/column,
    // since an exclusion can be introduced.

    init_state(Ms[E_NO_NO],al,ar,bl,br,
	       params->indel_score,0,params->indel_score,0);
    
    if (allow_exclusion) {
	init_state(Ms[E_X_NO], al,ar,bl,br, 0, params->exclusion_score, params->indel_score, 0);
	init_state(Ms[E_NO_X], al,ar,bl,br, params->indel_score, 0,    0, params->exclusion_score);
	init_state(Ms[E_X_X],  al,ar,bl,br, 0, params->exclusion_score, 0, params->exclusion_score);
	init_state(Ms[E_OP_NO],al,ar,bl,br, 0, 0,              params->indel_score, 0);
	init_state(Ms[E_NO_OP],al,ar,bl,br, params->indel_score, 0,    0, 0          );
	init_state(Ms[E_X_OP], al,ar,bl,br, 0, params->exclusion_score, 0, 0          );
	init_state(Ms[E_OP_X], al,ar,bl,br, 0, 0,              0, params->exclusion_score);   
    }

    FOR_IJ_IN_ARCMATCH
	Ms[E_NO_NO](i,j)=align_noex(Ms[E_NO_NO],al,bl,i,j);
    
    if (allow_exclusion) {
	int state;

	state=E_OP_NO;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i-1,j), Ms[E_NO_NO](i,j)); }
	    
	state=E_NO_OP;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i,j-1), Ms[E_NO_NO](i,j)); }

	state=E_NO_X;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(align_noex(Ms[state],al,bl,i,j),
						   Ms[E_NO_OP](i,j)+params->exclusion_score); }
	    	    
	state=E_OP_X;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i-1,j), Ms[E_NO_X](i,j)); }
	    
	state=E_X_NO;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(align_noex(Ms[state],al,bl,i,j),
						   Ms[E_OP_NO](i,j)+params->exclusion_score); }
	    
	state=E_X_OP;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i,j-1), Ms[E_X_NO](i,j)); }
	    
	state=E_X_X;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(align_noex(Ms[state],al,bl,i,j),
						   max(Ms[E_OP_X](i,j)+params->exclusion_score,
						       Ms[E_X_OP](i,j)+params->exclusion_score)); }
    }    
}

// compute the entries in the D matrix that
// can be computed from the matrix/matrices M
// for the subproblem al,bl,max_ar,max_br
//
// pre: M matrices are computed by a call to 
//      align_in_arcmatch(al,max_ar,bl,max_br,params->STRUCT_LOCAL)
//
void 
Aligner::fill_D_entries(int al, int bl,
			const BasePairs::LeftAdjList &adjlA,
			const BasePairs::LeftAdjList &adjlB,
			int max_ar, int max_br)
{
    for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	 arcA!=adjlA.end(); ++arcA) {
	for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
	     arcB!=adjlB.end(); ++arcB) {
	    //
	    // if right ends ar,br exceed the limits max_ar,max_br resp.
	    // the corresponding D entry is set to -infinity
	    // in order to dissalow the arc match completely
	    // Occurs only due to max-diff-am heuristic.
	    if (arcA->right()>max_ar || arcB->right()>max_br ) {
		std::cout << "Warning: -INFINITY"<<std::endl;
		D(arcA->idx(),arcB->idx()) = -100000; // ATTENTION: infinity constant (potential problem)
	    } else {
		if (!params->STRUCT_LOCAL)
		    D(arcA->idx(),arcB->idx()) = Ms[0](arcA->right()-1,arcB->right()-1);
		else {
		    // if we align structure local
		    // we need to determine the maximum of entries in Ms[k],
		    // where k is a non-open state
		    D(arcA->idx(),arcB->idx()) = INT_MIN;
		    for (int k=0; k<4; k++)
			D(arcA->idx(),arcB->idx())=max(D(arcA->idx(),arcB->idx()),
							Ms[k](arcA->right()-1,arcB->right()-1));
		}
		
		/*
		  if lonely pairs are disallowed,
		  we need to handle one extra case:
			  
		  there can be a stacked, enclosed arc, which is itself not
		  stacked to an enclosed arc

			  
		  e.g.
		  (((...........)))............
		  
		  
		  
		*/
		if (params->no_lonely_pair) {
		    if (bpsA.exists_arc(al+1,arcA->right()-1)
			&& bpsB.exists_arc(bl+1,arcB->right()-1)) {
			//
			const Arc &arcA2 = bpsA.arc(al+1,arcA->right()-1);
			const Arc &arcB2 = bpsB.arc(bl+1,arcB->right()-1);
			
			D(arcA->idx(),arcB->idx()) = 
			    max(D(arcA->idx(),arcB->idx()),
				D(arcA2.idx(),arcB2.idx())
				+arcA2.weight() + arcB2.weight()
				+tau(al+1,arcA->right()-1,bl+1,arcB->right()-1)
				);
		    }
		}
	    }
	}
    }
}


// compute all entries D
void Aligner::align_D() {
    // ------------------------------------------------------------
    // General workflow:
    //
    // for all combinations of left arc ends al and ar
    // 1.) determine for which arc-pairs the D entries can be computed   
    // in one run, 2.) call align_in_arcmatch 3.) call fill_D_entries
    // ------------------------------------------------------------

    // ------------------------------------------------------------
    // traverse the left ends al,bl of arcs in descending order
    //
    for (int al=r.get_endA(); al>r.get_startA()-1; al--) {
	const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);
	if (adjlA.size()>0) {
	    for (int bl=r.get_endB(); bl>r.get_startB()-1; bl--) {
		const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
		if (adjlB.size()>0) {
		    // ------------------------------------------------------------
		    // get maximal right ends of arcs with left ends al,bl 
		    // where max_diff_am conditions hold
		    //
		    int max_ar=al;
		    int max_br=bl;
		    if (params->max_diff_am>=0) {
			for (BasePairs::LeftAdjList::const_iterator
				 arcA = adjlA.begin();
			     arcA!=adjlA.end(); ++arcA) {
			    for (BasePairs::LeftAdjList::const_iterator
				     arcB = adjlB.begin();
				 arcB!=adjlB.end(); ++arcB) {
				int ar = arcA->right();
				int br = arcB->right();
				
				if (abs((ar-al)-(br-bl)) <= params->max_diff_am) {
				    max_ar=max(max_ar,ar);
				    max_br=max(max_br,br);
				}
			    }
			}
		    } else {
			max_ar=adjlA[adjlA.size()-1].right();
			max_br=adjlB[adjlB.size()-1].right();
		    }
		    
		    // ------------------------------------------------------------
		    // align under the maximal pair of arcs
		    //
		    align_in_arcmatch(al,max_ar,bl,max_br,params->STRUCT_LOCAL);

		    // ------------------------------------------------------------
		    // fill D matrix entries for the arc pairs beginning in al,bl
		    //
		    fill_D_entries(al,bl,adjlA,adjlB,max_ar,max_br);
		}
	    }
	}
    }

    D_created=true; // now the matrix D is built up
}


// align the top level in case of sequence local alignment
//
int Aligner::align_top_level_locally() {
    //std::cout << r << std::endl;
    
    ScoreMatrix &M=Ms[E_NO_NO];
    score_t max_score=0;
    init_state(M,r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1,0,0,0,0);
    
    for (uint i=r.get_startA(); i<=r.get_endA(); i++) {
	for (uint j=r.get_startB(); j<=r.get_endB(); j++) {
	    M(i,j) = 
		max( 0,
		     align_noex( M,r.get_startA()-1,r.get_startB()-1,i,j )
		     );
	    if (max_score < M(i,j)) {
		max_score=M(i,j);
		max_i = i;
		max_j = j;
	    }
	}
    }
    
    return max_score;
}


// special top level alignment for the scanning version
int Aligner::align_top_level_localB() {
    // std::cout <<"align local B " << r.get_startA() << " " << r.get_startB() << " "
    //           << r.get_endA() << " " << r.get_endB() << std::endl;
    
    ScoreMatrix &M=Ms[E_NO_NO];
    score_t max_score=INT_MIN;
    init_state(M,r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1,params->indel_score,0,0,0);
    
    for (uint i=r.get_startA(); i<=r.get_endA(); i++) {
	for (uint j=r.get_startB(); j<=r.get_endB(); j++) {
	    M(i,j) = 
		max( 0,
		     align_noex( M,r.get_startA()-1,r.get_startB()-1,i,j ) );
	    if (i==r.get_endA() && max_score < M(i,j)) {
		max_score=M(i,j);
		max_i = i;
		max_j = j;
	    }
	}
    }
    return max_score;
}


// compute the alignment score
score_t Aligner::align() {
    // ------------------------------------------------------------
    // computes D matrix and then
    // does the alignment on the top level
    // ------------------------------------------------------------
    
    if (!D_created) align_D();

    if (params->SEQU_LOCAL) {
	return align_top_level_locally();
    } else { // sequence global alignment
	//
	// align the subsequences that are specified by the restriction object
	//
	align_in_arcmatch(r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1,false);
	max_i=r.get_endA();
	max_j=r.get_endB();
	//
	return Ms[E_NO_NO](max_i,max_j);
    }
}

// ------------------------------------------------------------
// Aligner: traceback

void Aligner::trace_noex(int state,int al,int i,int bl,int j,bool tl) {
    ScoreMatrix &M=Ms[state];
    
    // determine where we get M(i,j) from
    
    // match
    if ( M(i,j) ==  M(i-1,j-1)+sigma(i,j) ) {
	trace_in_arcmatch(state,al,i-1,bl,j-1,tl);
	alignment.append(i,j);
	return;
    }
    // del
    if (M(i,j) == M(i-1,j)+params->indel_score) {
	trace_in_arcmatch(state,al,i-1,bl,j,tl);
	alignment.append(i,-1);
	return;
    }
    // ins
    if (M(i,j) == M(i,j-1)+params->indel_score) {
	trace_in_arcmatch(state,al,i,bl,j-1,tl);
	alignment.append(-1,j);
	return;
    }
	
    //  arc match
    const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i);
    const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j);
    
    for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin();
	 arcA!=adjlA.end() && arcA->left() > al; ++arcA) {
	//
	for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin();
	     arcB!=adjlB.end() && arcB->left() > bl ; ++arcB) {
	    //
	    if (params->no_lonely_pair
		&& bpsA.exists_arc(arcA->left()+1,i-1)
		&& bpsB.exists_arc(arcB->left()+1,j-1)) {
		//
		const Arc &arcA2 = bpsA.arc(arcA->left()+1,i-1);
		const Arc &arcB2 = bpsB.arc(arcB->left()+1,j-1);
		
		if ( M(i,j) == 
		     M(arcA->left()-1,arcB->left()-1) 
		     + D(arcA2.idx(),arcB2.idx())
		     + arcA->weight() + arcB->weight()
		     + tau(arcA->left(),arcB->left(),i,j)
		     + arcA2.weight() + arcB2.weight()
		     + tau(arcA->left()+1,arcB->left()+1,i-1,j-1)) {
		    //
		    // we know that (arcA,arcB) and (arcA2,arcB2) are arc-matches 
			
		    //cout << "Trace before arc match"<<endl;
		    trace_in_arcmatch(state,al,arcA->left()-1,bl,arcB->left()-1,tl);
		    
		    //cout<<endl;
		    //cout << "arcmatch "<<(arcA->left())<<","<<i<<";"<<(arcB->left())<<","<<j<<" :: "
		    //	 <<(arcA->w)<<" + "<<(arcB->w)<< " + " << tau(arcA->left(),arcB->left(),i,j)  <<endl;
		    
		    alignment.add_basepairA(arcA->left(),i);
		    alignment.add_basepairB(arcB->left(),j);
		    alignment.append(arcA->left(),arcB->left());
		    
		    alignment.add_basepairA(arcA->left()+1,i-1);
		    alignment.add_basepairB(arcB->left()+1,j-1);
		    alignment.append(arcA->left()+1,arcB->left()+1);
		    
		    //cout << "align in arcmatch (skipped AM due to noLP) "<<arcA->left()+1<<" "<<i-1<<" "<<arcB->left()+1<<" "<<j-1<<endl;
		    align_in_arcmatch(arcA->left()+1,i-1,arcB->left()+1,j-1,params->STRUCT_LOCAL); // this fills matrices Ms[k]
		    
		    //cout << arcA2.idx() << " " <<arcB2.idx() << endl;
		    
		    if (!params->STRUCT_LOCAL) {
			trace_in_arcmatch(0,arcA->left()+1,i-2,arcB->left()+1,j-2,false);
		    } else {
			// there are two cases: the optimal value D(arcA2.idx(),arcB2.idx()) can
			//   1.) stem from matrices Ms[k]
			//   2.) stem from an inner arc-match "D(arcA3.idx(),arcB3.idx())"
			
			uint k;
			for (k=0; k<4; k++) {
			    if (D(arcA2.idx(),arcB2.idx()) == Ms[k](i-2,j-2)) {
				//cout << "trace in arcmatch "<<k<<" "<<arcA->left()<<" "<<i<<endl;
				trace_in_arcmatch(k,arcA->left()+1,i-2,arcB->left()+1,j-2,false);
				break;
			    }
			    //cout <<"k:"<<k<<" "<<D(arcA2.idx(),arcB2.idx())<<" "<<Ms[k](i-2,j-2)<<endl;
			}
			// was ist, wenn das scheitert???
			// !!!!!!!!!! DANN ist D wieder gestackt gewesen: IST DAS JETZT KORREKT ???????????
			if (k==4) {
			    //cout << arcA2.idx() << " " <<arcB2.idx()<< " gestackt" << endl;
			    trace_in_arcmatch(0,arcA->left()+1,i-2,arcB->left()+1,j-2,false);
			}
		    }
		    
		    alignment.append(i-1,j-1);
		    alignment.append(i,j);
		    return;
		}
	    }
	    
	    // here, either lonely pairs are allowed or there is no inner base pair
	    if ( M(i,j) == 
		 M(arcA->left()-1,arcB->left()-1) 
		 + D(arcA->idx(),arcB->idx())
		 + arcA->weight() + arcB->weight()
		 + tau(arcA->left(),arcB->left(),i,j)) {
		//
		trace_in_arcmatch(state,al,arcA->left()-1,bl,arcB->left()-1,tl);
		
		//cout << "arcmatch "<<(arcA->left())<<","<<i<<";"<<(arcB->left())<<","<<j<<" :: "
		//      <<(arcA->w)<<" + "<<(arcB->w)<< " + " << tau(arcA->left(),arcB->left(),i,j)  <<endl;
		
		alignment.add_basepairA(arcA->left(),i);
		alignment.add_basepairB(arcB->left(),j);
		alignment.append(arcA->left(),arcB->left());
		
		align_in_arcmatch(arcA->left(),i, arcB->left(),j,
				  params->STRUCT_LOCAL);
		
		if (!params->STRUCT_LOCAL) {
		    trace_in_arcmatch(0,arcA->left(),i-1,arcB->left(),j-1,false);
		} else {
		    for (uint k=0; k<4; k++) {
			if (D(arcA->idx(),arcB->idx()) == Ms[k](i-1,j-1)) {
			    trace_in_arcmatch(k,arcA->left(),i-1,arcB->left(),j-1,false);
			    break;
			}
		    }
		}
		
		alignment.append(i,j);
		return;
	    }
	}
    }
}

void Aligner::trace_in_arcmatch(int state,int al,int i,int bl,int j,bool tl) {
    //pre: M matrices for arc computed
    ScoreMatrix &M=Ms[state];

    /*string state_text[]={"E_NO_NO", "E_X_NO", "E_NO_X", "E_X_X",
      "E_OP_NO", "E_NO_OP", "E_OP_X", "E_X_OP"};
      cout << "trace_in_arcmatch "<<state_text[state]<<" al:"<<al<<" i:"<<i
      <<" bl:"<<bl<<" j:"<<j<<" :: "<< M(i,j) <<endl;
    */
    
    // terminate traceback if
    // * trace on toplevel (tl)
    // * entry 0
    if (tl && M(i,j)==0) {
	return;
    }
    if (i<=al) {
	if (state==E_NO_NO
	    || state==E_OP_NO
	    || state==E_X_NO) {
	    for (int k=bl+1;k<=j;k++)
		alignment.append(-1,k);
	} else {
	    //cout << "exclude B "<<state<<" "<< bl<<" - "<<j<<endl;
	}
	return;
    }

    if (j<=bl) {
	if (state==E_NO_NO
	    || state==E_NO_OP
	    || state==E_NO_X) {
	    for (int k=al+1;k<=i;k++) {
		alignment.append(k,-1);
	    }
	} else {
	    //cout << "exclude A "<<state<<" "<< al<<" - "<<i<<endl;
	}
	return;
    }


    switch(state) {
    case E_NO_NO:
	trace_noex(state,al,i,bl,j,tl);
	break;
    case E_OP_NO:
	if (M(i,j) == M(i-1,j)) {
	    // cout << "exclude A "<<i<<endl;
	    trace_in_arcmatch(state,al,i-1,bl,j,tl);
	} else if (M(i,j) == Ms[E_NO_NO](i,j))
	    trace_in_arcmatch(E_NO_NO,al,i,bl,j,tl);
	break;
    case E_NO_OP:
	if (M(i,j) == M(i,j-1)) {
	    // cout << "exclude B "<<j<<endl;
	    trace_in_arcmatch(state,al,i,bl,j-1,tl);
	} else if (M(i,j) == Ms[E_NO_NO](i,j))
	    trace_in_arcmatch(E_NO_NO,al,i,bl,j,tl);
	break;
    case E_NO_X:
	if (M(i,j) == Ms[E_NO_OP](i,j)+params->exclusion_score) {
	    trace_in_arcmatch(E_NO_OP,al,i,bl,j,tl);
	} else
	    trace_noex(state,al,i,bl,j,tl);
	break;
    case E_OP_X:
	if (M(i,j) == M(i-1,j)) {
	    //cout << "exclude A "<<i<<endl;
	    trace_in_arcmatch(state,al,i-1,bl,j,tl);
	} else if (M(i,j) == Ms[E_NO_X](i,j))
	    trace_in_arcmatch(E_NO_X,al,i,bl,j,tl);
	break;
    case E_X_NO:
	if (M(i,j) == Ms[E_OP_NO](i,j)+params->exclusion_score) {
	    trace_in_arcmatch(E_OP_NO,al,i,bl,j,tl);
	} else
	    trace_noex(state,al,i,bl,j,tl);
	break;
    case E_X_OP:
	if (M(i,j) == M(i,j-1)) {
	    // cout << "exclude B "<<j<<endl;
	    trace_in_arcmatch(state,al,i,bl,j-1,tl);
	} else if (M(i,j) == Ms[E_X_NO](i,j))
	    trace_in_arcmatch(E_X_NO,al,i,bl,j,tl);
	break;
    case E_X_X:
	if (M(i,j) == Ms[E_OP_X](i,j)+params->exclusion_score) {
	    trace_in_arcmatch(E_OP_X,al,i,bl,j,tl); 
	} else if (M(i,j) == Ms[E_X_OP](i,j)+params->exclusion_score) {
	    trace_in_arcmatch(E_X_OP,al,i,bl,j,tl);
	} else
	    trace_noex(state,al,i,bl,j,tl);
	break;
    }
}


void Aligner::trace() {
    // pre: last call align_in_arcmatch(r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1);
    //      or align_top_level_locally for SEQU_LOCAL alignment
    
    // reset the alignment strings (to empty strings)
    // such that they can be written again during the trace
    alignment.clear();
    
    trace_in_arcmatch(E_NO_NO,r.get_startA()-1,max_i,r.get_startB()-1,max_j,params->SEQU_LOCAL);
}

