#ifndef MATRICES_HH
#define MATRICES_HH

/*
  Define classes for the dynamic programming
  matrices.

  The structures support offsets and moving
  of the offset, while maintaining the entries
  in the overlapping sub-matrix
*/

template <class elem_t>
class Matrix {
public:
    typedef elem_t val_t;
protected:
    size_t xdim_;
    size_t ydim_;
    std::vector<score_t> mat_;

    size_t addr(size_t i, size_t j) const {
	assert(0<=i && i<this->xdim_);
	assert(0<=j && j<this->ydim_);
	return i*ydim_+j;
    }

public:
    Matrix() : mat_(),xdim_(0),ydim_(0) {
    }
        
    void
    resize(size_t xdim, size_t ydim) {
	xdim_=xdim;
	ydim_=ydim;
	mat_.resize(xdim*ydim);
    }

    const elem_t & operator() (size_t i,size_t j) const {
	return mat_[addr(i,j)];
    }

    elem_t & operator() (size_t i,size_t j) {
	return mat_[addr(i,j)];
    }
};

// ----------------------------------------
//! simple matrix class with offset
template <class elem_t>
class OMatrix : public Matrix<elem_t> {
protected:
    size_t off_;
    size_t xoff_;
    size_t yoff_;

    size_t addr(size_t i, size_t j) const {
	assert(xoff_<=i && i<xoff_+this->xdim_);
	assert(yoff_<=j && j<yoff_+this->ydim_);
	return i*this->xdim_ + j - off_;
    }
    
public:
    OMatrix() : Matrix<elem_t>(0) {
    }
        
    void
    resize(size_t xdim, size_t ydim, size_t xoff=0, size_t yoff=0) {
	xoff_=xoff;
	yoff_=yoff;
	off_=xoff*xdim+yoff;
	this->xdim_=xdim;
	this->ydim_=ydim;
	this->mat_.resize(xdim*ydim);
    }

    const elem_t & operator() (size_t i,size_t j) const {
	return this->mat_[addr(i,j)];
    }

    elem_t & operator() (size_t i,size_t j) {
	return this->mat_[addr(i,j)];
    }
};


// ----------------------------------------
//! matrix class with rotation
template <class elem_t>
class RMatrix: public Matrix<elem_t> {

protected:
    size_t xrot_;
    size_t yrot_;

    size_t rot(size_t x, size_t r, size_t d) {
	assert(r<d);
	return (x+d-r)%d;
    }
    
    size_t addr(size_t i, size_t j) const {
	assert(xrot_<=i && i<xrot_+this->xdim_);
	assert(yrot_<=j && j<yrot_+this->ydim_);
	return rot(i,xrot_,this->xdim_)*this->xdim_ + rot(j,yrot_,this->ydim_);
    }

public:    
    RMatrix() : Matrix<elem_t>(0) {
    }
        
    void
    resize(size_t xdim, size_t ydim, size_t xrot=0, size_t yrot=0) {
	xrot_=xrot;
	yrot_=yrot;
	this->xdim_=xdim;
	this->ydim_=ydim;
	this->mat_.resize(xdim*ydim);
    }

    void move(size_t xrot, size_t yrot) {
	xrot=xrot_;
	yrot=yrot_;
    }
    
    const elem_t & operator() (size_t i,size_t j) const {
	return this->mat_[addr(i,j)];
    }

    elem_t & operator() (size_t i,size_t j) {
	return this->mat_[addr(i,j)];
    }
};


#endif // MATRICES_HH
